# LOAD DATA
import pandas as pd

{% if dataset.training_data_path.endswith(".pkl") %}
train_dataset = pd.read_pickle("./training.pkl")
{% else %}
train_dataset = pd.read_csv("./training.csv", encoding="{{ dataset.csv_encoding }}", delimiter="{{ dataset.csv_delimiter }}")
{% endif %}
